import sys
from pathlib import Path

import pytest

ROOT = Path(__file__).resolve().parents[1]
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))

from device_smi import Device
from device_smi.base import GPU, BaseMetrics, GPUDevice, Pcie


@pytest.fixture()
def gpu_device():
    try:
        return Device("gpu")
    except Exception as exc:  # pragma: no cover - environment dependent
        pytest.skip(f"GPU device not available or tooling missing: {exc}")


def test_gpu_device_basic_attributes(gpu_device):
    assert gpu_device.type == "gpu"
    assert isinstance(gpu_device.device, GPUDevice)
    assert isinstance(gpu_device.model, str) and gpu_device.model
    assert isinstance(gpu_device.vendor, str) and gpu_device.vendor
    assert isinstance(gpu_device.memory_total, int) and gpu_device.memory_total > 0


def test_gpu_device_optional_components(gpu_device):
    if gpu_device.pcie is not None:
        assert isinstance(gpu_device.pcie, Pcie)
        assert isinstance(gpu_device.pcie.gen, int)
        assert isinstance(gpu_device.pcie.speed, int)
        assert isinstance(gpu_device.pcie.id, str) and gpu_device.pcie.id
    if gpu_device.gpu is not None:
        assert isinstance(gpu_device.gpu, GPU)
        assert isinstance(gpu_device.gpu.driver, str) and gpu_device.gpu.driver
        assert isinstance(gpu_device.gpu.firmware, str)


@pytest.mark.parametrize("fast", [False, True])
def test_gpu_metrics(gpu_device, fast):
    try:
        metrics = gpu_device.metrics(fast=fast)
    except RuntimeError as exc:  # pragma: no cover - depends on vendor tooling
        pytest.skip(f"GPU metrics unavailable: {exc}")

    assert isinstance(metrics, BaseMetrics)
    assert metrics.memory_used >= 0
    assert metrics.memory_process >= 0
    assert metrics.utilization >= 0



import re
import subprocess
from abc import abstractmethod
from typing import Optional

INSTALLATION_HINTS = {
    "lspci": "`lspci` is not installed, you can install it via `sudo apt install pciutils`.",
    "nvidia-smi": "`nvidia-smi` is not installed. You need to install NVIDIA driver support binaries by `sudo apt install nvidia-utils-<NVIDIA-DRIVER-VERSION>`",
    "powershell": "`PowerShell` is not installed. Please follow the instructions at `https://learn.microsoft.com/en-us/powershell/scripting/install/installing-powershell`",
    "xpu-smi": "`xpu-smi` is not installed. Please follow the instructions at  https://github.com/intel/xpumanager/blob/master/doc/smi_install_guide.md`",
    "amd-smi": "`amd-smi` is not installed. Please follow the instructions at  `https://rocm.docs.amd.com/projects/amdsmi/en/latest/install/install.html`",
}

class BaseDevice:
    # Indicates whether the fast metrics path is identical to the regular metrics call.
    fast_metrics_same_as_slow = False

    def __init__(self, cls, type: str):
        cls.type = type

    @abstractmethod
    def metrics(self):
        pass

    def __str__(self):
        return str(self.__dict__)

    def to_dict(self, text, split: str = ":"):
        return {k.strip(): v.strip() for k, v in (line.split(split, 1) for line in text.splitlines() if split in line)}


class GPUDevice(BaseDevice):
    def __init__(self, cls, index):
        super().__init__(cls, "gpu")
        self.index = index

    @abstractmethod
    def metrics(self):
        pass


class BaseMetrics:
    def __init__(
        self,
        memory_used: int = 0,
        memory_process: int = 0,
        utilization: float = 0.0,
    ):
        self.memory_used = memory_used
        self.memory_process = memory_process
        self.utilization = max(0.0, utilization)

    def __str__(self):
        return str(self.__dict__)


class Pcie:
    def __init__(self, gen: int, speed: int, id: str):
        self.gen = gen
        self.speed = speed
        self.id = id

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return self.__str__()


class GPU:
    def __init__(self, driver: str, firmware: str):
        self.driver = driver
        self.firmware = firmware

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return self.__str__()


def _run(args, line_start: Optional[str] = None, seperator: str=None): # -> str | list[str] disable type hint, because solaris test is using python 3.7 which doesn't support | usage
    try:
        result = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    except FileNotFoundError:
        install_hint = INSTALLATION_HINTS.get(args[0], f"Command not found: `{args[0]}`, please check if it was installed.")
        raise RuntimeError(install_hint)

    if result.returncode != 0 or result.stderr.strip() != "":
        raise RuntimeError(result.stderr)

    result = result.stdout.strip()
    result = re.sub(r'\n+', '\n', result) # remove consecutive \n
    if line_start:
        return " ".join([line for line in result.splitlines() if line.strip().startswith(line_start)])
    if seperator:
        return [l.strip() for l in result.split(seperator)]
    return result



import os
import warnings

from .base import GPU, BaseMetrics, GPUDevice, Pcie, _run


class NvidiaGPUMetrics(BaseMetrics):
    pass


class NvidiaDevice(GPUDevice):
    fast_metrics_same_as_slow = False

    def __init__(self, cls, index):
        super().__init__(cls, index)
        self.gpu_id = self._get_gpu_id()

        try:
            args = [
                "nvidia-smi",
                f"--id={self.gpu_id}",
                "--query-gpu="
                "name,"
                "memory.total,"
                "pci.bus_id,"
                "pcie.link.gen.max,"
                "pcie.link.gen.current,"
                "driver_version",
                "--format=csv,noheader,nounits",
            ]

            result = _run(args=args, seperator="\n")

            model, total_memory, pci_bus_id, pcie_gen, pcie_width, driver = (result[0].split(", "))

            result = _run(args=["nvidia-smi", "-q", "-i", f"{self.gpu_id}"], seperator="\n")
            firmware = " ".join([line.split(":", 1)[1].strip() for line in result if "VBIOS" in line])

            if model.lower().startswith("nvidia"):
                model = model[len("nvidia"):]

            compute_cap = (
                _run(["nvidia-smi", "--format=csv", "--query-gpu=compute_cap", "-i", f"{self.gpu_id}"])
                .removeprefix("compute_cap\n")
            )

            cls.model = model.strip().lower()
            cls.memory_total = int(total_memory) * 1024 * 1024  # bytes
            cls.vendor = "nvidia"
            cls.features = [compute_cap]
            cls.pcie = Pcie(gen=int(pcie_gen), speed=int(pcie_width), id=pci_bus_id)
            cls.gpu = GPU(driver=driver, firmware=firmware)
        except FileNotFoundError:
            raise FileNotFoundError()
        except Exception as e:
            raise e

    def _get_gpu_id(self):
        gpu_count = len(_run(["nvidia-smi", "--list-gpus"]).splitlines())
        cudas = os.environ.get("CUDA_VISIBLE_DEVICES", "")
        cuda_list = cudas.split(",") if cudas else []
        if gpu_count > 0 and os.environ.get("CUDA_DEVICE_ORDER", "") != "PCI_BUS_ID":
            warnings.warn("Detected different devices in the system. Please make sure to set `CUDA_DEVICE_ORDER=PCI_BUS_ID` to avoid unexpected behavior.", RuntimeWarning, 2)
        if cuda_list and len(cuda_list) > self.index:
            return cuda_list[self.index]
        else:
            return str(self.index)

    def metrics(self):
        try:
            args = ["nvidia-smi", f"--id={self.gpu_id}", "--query-gpu=memory.used,utilization.gpu", "--format=csv,noheader,nounits"]
            used_memory, utilization = _run(args=args, seperator="\n")[0].split(", ")

            return NvidiaGPUMetrics(
                memory_used=int(used_memory) * 1024 * 1024,  # bytes
                memory_process=0,  # Bytes, TODO, get this
                utilization=float(utilization),
            )

        except FileNotFoundError:
            raise FileNotFoundError(
                "The 'nvidia-smi' command was not found. Please ensure that the 'nvidia-utils' package is installed."
            )
        except Exception as e:
            raise e



import os
import platform
import re

from .base import BaseDevice, BaseMetrics, _run


class CPUMetrics(BaseMetrics):
    pass


class CPUDevice(BaseDevice):
    fast_metrics_same_as_slow = False

    def __init__(self, cls):
        super().__init__(cls, "cpu")

        model = "Unknown Model"
        vendor = "Unknown vendor"
        flags = set()

        if platform.system().lower() == "windows":
            command_result = _run(["wmic", "cpu", "get", "manufacturer,name,numberofcores,numberoflogicalprocessors", "/format:csv"]).strip()
            result = command_result.split("\n")[1].split(",")

            cpu_count = command_result.count('\n')
            model = result[2].strip()
            cpu_cores = int(result[3])
            cpu_threads = int(result[4])
            vendor = result[1].strip()

            command_result = _run(["wmic", "os", "get", "TotalVisibleMemorySize", "/Value", "/format:csv"]).strip()
            result = command_result.split("\n")[1].split(",")

            mem_total = int(result[1])
        elif platform.system().lower() == 'darwin':
            model = (_run(["sysctl", "-n", "machdep.cpu.brand_string"]).replace("Apple", "").strip())
            try:
                vendor = (_run(["sysctl", "-n", "machdep.cpu.vendor"]))
            except BaseException:
                vendor = "apple"

            sysctl_info = self.to_dict(_run(["sysctl", "-a"]))
            cpu_count = 1
            cpu_cores = int(sysctl_info["hw.physicalcpu"])
            cpu_threads = int(sysctl_info["hw.logicalcpu"])

            mem_total = int(_run(["sysctl", "-n", "hw.memsize"]))

            try:
                features = sysctl_info["machdep.cpu.features"].splitlines()
            except Exception:
                features = []

            flags = set(features)
        elif os.name == 'posix':
            try:
                with open("/proc/cpuinfo", "r") as f:
                    lines = f.readlines()
                    for line in lines:
                        if line.startswith("flags"):
                            flags.update(line.strip().split(":")[1].split())
                        if line.startswith("model name"):
                            model = line.split(":")[1].strip()
                        elif line.startswith("vendor_id"):
                            vendor = line.split(":")[1].strip()
            except FileNotFoundError:
                model = platform.processor()
                vendor = platform.uname().system

            cpu_info = self.to_dict(_run(['lscpu']))

            cpu_count = int(cpu_info["Socket(s)"])
            cpu_cores_per_socket = int(cpu_info["Core(s) per socket"])
            cpu_cores = cpu_count * cpu_cores_per_socket
            cpu_threads = int(cpu_info["CPU(s)"])

            with open("/proc/meminfo", "r") as f:
                lines = f.readlines()
                mem_total = 0
                for line in lines:
                    if line.startswith("MemTotal:"):
                        mem_total = int(line.split()[1]) * 1024
                        break
        else:
            print("not support")

        model = " ".join(i for i in model.lower().split() if not any(x in i for x in ["ghz", "cpu", "(r)", "(tm)", "intel", "amd", "core", "processor", "@"]))
        cls.model = model

        if "intel" in vendor.lower():
            vendor = "intel"
        elif "amd" in vendor.lower():
            vendor = "amd"
        cls.vendor = vendor.lower().replace("authentic", "")
        cls.memory_total = mem_total  # Bytes
        self.memory_total = mem_total  # Bytes
        cls.count = cpu_count
        cls.cores = cpu_cores
        cls.threads = cpu_threads
        cls.features = sorted({f.lower() for f in flags})

    def _utilization(self):
        # check if is macOS
        if platform.system().lower() == "darwin":
            output = _run(["top", "-l", "1", "-stats", "cpu"])

            # CPU usage: 7.61% user, 15.23% sys, 77.15% idle
            for line in output.splitlines():
                if line.startswith("CPU usage"):
                    parts = line.split()
                    user_time = float(parts[2].strip("%"))
                    sys_time = float(parts[4].strip("%"))
                    idle_time = float(parts[6].strip("%"))
                    total_time = user_time + sys_time + idle_time
                    return total_time, idle_time
        else:
            with open("/proc/stat", "r") as f:
                lines = f.readlines()
                for line in lines:
                    if line.startswith("cpu "):
                        parts = line.split()
                        total_time = sum(int(part) for part in parts[1:])
                        idle_time = int(parts[4])
                        return total_time, idle_time

    def metrics(self):
        if platform.system().lower() == "windows":
            if platform.system().lower() == "windows":
                command_result = _run(["wmic", "cpu", "get", "loadpercentage"]).strip()
                try:
                    result = command_result.split("\n")[1].split(",")
                    utilization = int(result[0])
                except BaseException as e:
                    print("error occurred, command_result: ")
                    print(f"{command_result}")
                    print("------------")
                    raise e

                try:
                    command_result = _run(["wmic", "os", "get", "FreePhysicalMemory"]).strip()
                    result = command_result.split("\n")[1].split(",")
                    memory_used = int(result[0])
                except BaseException as e:
                    print("error occurred, command_result: ")
                    print(f"{command_result}")
                    print("------------")
                    raise e
                return CPUMetrics(
                    memory_used=memory_used,  # bytes
                    memory_process=0,  # bytes
                    utilization=utilization,
                )

        total_time_1, idle_time_1 = self._utilization()
        # read CPU status second time here, read too quickly will get inaccurate results
        total_time_2, idle_time_2 = self._utilization()

        total_diff = total_time_2 - total_time_1
        idle_diff = idle_time_2 - idle_time_1

        # total_diff might be 0
        if total_diff <= 0:
            utilization = 0
        else:
            if platform.system().lower() == "darwin":
                utilization = idle_time_2 - idle_time_1
            else:
                utilization = (1 - (idle_diff / total_diff)) * 100

        if platform.system().lower() == "darwin":
            available_mem = _run(["vm_stat"]).replace(".", "").lower()

            result = self.to_dict(available_mem)

            available_mem = available_mem.splitlines()
            page_size = int(re.findall(r'\d+', available_mem[0])[0])

            free_pages = int(result["pages free"])

            mem_free = free_pages * page_size
        else:
            with open("/proc/meminfo", "r") as f:
                lines = f.readlines()
                mem_free = 0
                for line in lines:
                    if line.startswith("MemAvailable:"):
                        mem_free = int(line.split()[1]) * 1024
                        break

        memory_used = self.memory_total - mem_free

        process_id = os.getpid()
        if platform.system().lower() == "darwin":
            result = _run(["ps", "-p", str(process_id), "-o", "rss="])
            memory_current_process = int(result) * 1024
        else:
            with open(f"/proc/{process_id}/status", "r") as f:
                lines = f.readlines()
                memory_current_process = 0
                for line in lines:
                    if line.startswith("VmRSS:"):
                        memory_current_process = int(line.split()[1]) * 1024
                        break

        return CPUMetrics(
            memory_used=memory_used,  # bytes
            memory_process=memory_current_process,  # bytes
            utilization=utilization,
        )



import platform
import threading
import warnings
from typing import Optional

from .amd import AMDDevice
from .apple import AppleDevice
from .base import _run
from .cpu import CPUDevice
from .intel import IntelDevice
from .nvidia import NvidiaDevice
from .os import OSDevice

IS_ROCM = False
try:
    import torch

    HAS_TORCH = True
    if torch.version.hip is not None:
        IS_ROCM = True
except BaseException:
    HAS_TORCH = False


class Device:
    def __init__(self, device, *, fast_metrics_interval: float = 0.200):
        # init attribute first to avoid IDE not attr warning
        # CPU/GPU Device
        self.memory_total = None
        self.type = None
        self.features = []
        self.vendor = None
        self.model = None
        self.device = None
        # OS Device
        self.arch = None
        self.version = None
        self.name = None
        self._fast_metrics_thread: Optional[threading.Thread] = None
        self._fast_metrics_cache = None
        self._fast_metrics_error: Optional[RuntimeError] = None
        self._fast_metrics_lock = threading.Lock()
        self._metrics_lock = threading.Lock()
        self._fast_metrics_stop_event: Optional[threading.Event] = None
        self._fast_metrics_interval = self._validate_fast_metrics_interval(fast_metrics_interval)
        self._fast_metrics_same_as_slow = False
        if HAS_TORCH and isinstance(device, torch.device):
            device_type = device.type.lower()
            device_index = device.index
        elif f"{device}".lower() == "os":
            self.device = OSDevice(self)
            self._configure_fast_metrics()
            return
        else:
            d = f"{device}".lower()
            if ":" in d:
                type, index = d.split(":")
                device_type = type
                device_index = int(index)
            else:
                device_type = d
                device_index = 0

        self.pcie = None
        self.gpu = None

        if device_type == "cpu":
            self.device = CPUDevice(self)
        elif device_type == "xpu":
            self.device = IntelDevice(self, device_index)
        elif device_type == "rocm" or IS_ROCM:
            self.device = AMDDevice(self, device_index)
        elif device_type == "cuda" and not IS_ROCM:
            self.device = NvidiaDevice(self, device_index)
        elif device_type == "gpu":
            if platform.system().lower() == "darwin":
                if platform.machine() == "x86_64":
                    raise Exception("Not supported for macOS on Intel chips.")

                self.device = AppleDevice(self, device_index)
            else:
                if platform.system().lower() == "windows":
                    import os
                    import shutil

                    if not shutil.which("powershell.exe"):
                        psdir = os.path.join(
                            os.environ.get("SystemRoot", r"C:\Windows"),
                            "System32",
                            "WindowsPowerShell",
                            "v1.0",
                        )
                        os.environ["PATH"] = os.environ.get("PATH", "") + ";" + psdir

                    for d in ["NVIDIA", "AMD", "INTEL"]:
                        result = (
                            _run(
                                [
                                    "powershell",
                                    "-Command",
                                    "Get-CimInstance",
                                    "Win32_VideoController",
                                    "-Filter",
                                    f"\"Name like '%{d}%'\"",
                                ]
                            )
                            .lower()
                            .splitlines()
                        )
                        if result:
                            if d == "INTEL":
                                self.device = IntelDevice(self, device_index)
                            elif d == "AMD":
                                self.device = AMDDevice(self, device_index)
                            else:
                                self.device = NvidiaDevice(self, device_index)
                            break
                else:
                    result = _run(["lspci"]).lower().splitlines()
                    result = "\n".join(
                        [
                            line
                            for line in result
                            if any(
                                keyword.lower() in line.lower()
                                for keyword in ["vga", "3d", "display"]
                            )
                        ]
                    ).lower()
                    if "nvidia" in result:
                        self.device = NvidiaDevice(self, device_index)
                    elif "amd" in result:
                        self.device = AMDDevice(self, device_index)
                    elif "intel" in result:
                        self.device = IntelDevice(self, device_index)
            if not self.device:
                raise ValueError(f"Unable to find requested device: {device}")
        else:
            raise Exception(f"The device {device_type} is not supported")

        self._configure_fast_metrics()

    def info(self):
        warnings.warn(
            "info() method is deprecated and will be removed in next release.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self

    def memory_used(self) -> int:
        return self.metrics().memory_used

    def utilization(self) -> float:
        return self.metrics().utilization

    def metrics(self, fast: bool = False):
        if not fast or self._fast_metrics_same_as_slow:
            metrics = self._collect_metrics()
            self._update_fast_cache(metrics)
            return metrics

        metrics, error = self._get_cached_metrics()
        if error is not None:
            raise error
        if metrics is not None:
            return metrics

        metrics = self._collect_metrics()
        self._update_fast_cache(metrics)
        return metrics

    def close(self):
        self._stop_fast_metrics_worker()

    def __str__(self):
        return str(
            {k: v for k, v in self.__dict__.items() if k != "device" and v is not None}
        )

    def __del__(self):  # pragma: no cover - defensive cleanup only
        try:
            self._stop_fast_metrics_worker()
        except Exception:
            pass

    @staticmethod
    def _validate_fast_metrics_interval(value: float) -> float:
        try:
            interval = float(value)
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive branch
            raise ValueError("fast_metrics_interval must be a positive number") from exc
        if interval <= 0:
            raise ValueError("fast_metrics_interval must be greater than 0")
        return interval

    def _configure_fast_metrics(self):
        if not self.device:
            return
        self._fast_metrics_same_as_slow = bool(
            getattr(self.device, "fast_metrics_same_as_slow", False)
        )
        if self._fast_metrics_same_as_slow:
            return

        self._fast_metrics_stop_event = threading.Event()
        thread_name = f"device-smi-metrics-{self.type or 'unknown'}"
        self._fast_metrics_thread = threading.Thread(
            target=self._fast_metrics_worker,
            name=thread_name,
            daemon=True,
        )
        self._fast_metrics_thread.start()

    def _fast_metrics_worker(self):
        stop_event = self._fast_metrics_stop_event
        assert stop_event is not None
        while not stop_event.is_set():
            try:
                metrics = self._collect_metrics()
            except Exception as exc:
                with self._fast_metrics_lock:
                    self._fast_metrics_error = RuntimeError(str(exc))
                if stop_event.wait(self._fast_metrics_interval):
                    break
                continue

            self._update_fast_cache(metrics)
            if stop_event.wait(self._fast_metrics_interval):
                break

    def _collect_metrics(self):
        with self._metrics_lock:
            return self.device.metrics()

    def _update_fast_cache(self, metrics):
        if self._fast_metrics_same_as_slow:
            return
        with self._fast_metrics_lock:
            self._fast_metrics_cache = metrics
            self._fast_metrics_error = None

    def _get_cached_metrics(self):
        with self._fast_metrics_lock:
            return self._fast_metrics_cache, self._fast_metrics_error

    def _stop_fast_metrics_worker(self):
        if self._fast_metrics_stop_event is None:
            return
        # signal stop
        self._fast_metrics_stop_event.set()
        thread = self._fast_metrics_thread

        if thread and thread.is_alive():
            thread.join(timeout=0.5)

        # only clear references after the worker is finished
        self._fast_metrics_thread = None
        self._fast_metrics_stop_event = None



import re
import sys
from pathlib import Path

import pytest

ROOT = Path(__file__).resolve().parents[1]
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))

from device_smi import Device
from device_smi.os import OSDevice


@pytest.fixture()
def os_device():
    device = Device("os")
    try:
        yield device
    finally:
        device.close()


def test_os_device_basic_attributes(os_device):
    assert os_device.type == "os"
    assert isinstance(os_device.device, OSDevice)
    assert isinstance(os_device.name, str) and os_device.name

    version = os_device.version
    assert isinstance(version, str) and version
    if version[0].isdigit():
        assert re.match(r"[0-9]+(\.[0-9]+)*", version)

    assert isinstance(os_device.kernel, str) and os_device.kernel
    assert isinstance(os_device.arch, str) and os_device.arch
    assert os_device.arch == os_device.arch.lower()
    assert re.match(r"[0-9a-z_\-]+", os_device.arch)


def test_os_device_metrics(os_device):
    assert os_device.metrics() is None
    assert os_device.device.fast_metrics_same_as_slow is True
    with pytest.raises(AttributeError):
        os_device.memory_used()



import os
import platform
import re

from .base import BaseDevice, BaseMetrics, _run


class OSMetrics(BaseMetrics):
    pass


class OSDevice(BaseDevice):
    fast_metrics_same_as_slow = True

    def __init__(self, cls):
        super().__init__(cls, "os")

        if platform.system().lower() == "linux" or platform.system().lower() == "freebsd" or platform.system().lower() == "solaris" or platform.system().lower() == "sunos":
            release_info = self.to_dict(_run(["cat", "/etc/os-release"]).replace("\"", "").lower(), "=")
            cls.name = release_info["name"].replace("oracle", "").replace("gnu/linux", "").strip()

            cls.version = release_info["version_id"]
            match = re.match(r"(\d+\.\d+)", cls.version)
            if match:
                cls.version = match.group(1)

            cls.kernel, cls.arch = _run(["uname", "-mr"]).lower().split()
        elif platform.system().lower() == "darwin":
            release_info = self.to_dict(_run(["sw_vers"]).lower())
            cls.name = release_info["productname"]
            cls.version = release_info["productversion"]
            cls.kernel, cls.arch = _run(["uname", "-mr"]).lower().split()
        elif platform.system().lower() == "windows":
            cls.version = _run(["wmic", "os", "get", "caption", "/format:csv"], seperator="\n")[1].split(",")[1].lower().removeprefix("microsoft windows").strip()
            cls.name = "windows"
            cls.arch = os.environ.get("PROCESSOR_ARCHITECTURE").lower()

            cls.kernel = _run(["cmd", "/c", "ver"])
            match = re.search(r'(\d+\.\d+\.\d+\.\d+)', cls.kernel)
            if match:
                cls.kernel = match.group(1)
        else:
            cls.name = platform.system().lower()
            cls.version = platform.version().lower()
            cls.arch = platform.architecture()[0].lower().strip()

        if cls.arch in ["amd64", "x64"]:
            cls.arch = "x86_64"
        if cls.arch in ["i386", "i86pc"]:
            cls.arch = "x86"
        if cls.arch in ["arm64"]:
            cls.arch = "aarch64"

    def metrics(self):
        pass



import sys
from pathlib import Path

import pytest

ROOT = Path(__file__).resolve().parents[1]
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))

from device_smi import Device
from device_smi.base import BaseMetrics
from device_smi.cpu import CPUDevice, CPUMetrics


@pytest.fixture()
def cpu_device():
    return Device("cpu")


def test_cpu_device_basic_attributes(cpu_device):
    banned_tokens = {"ghz", "cpu", "(r)", "(tm)", "intel", "amd", "core", "processor", "@"}

    assert cpu_device.type == "cpu"
    assert isinstance(cpu_device.device, CPUDevice)
    assert isinstance(cpu_device.model, str) and cpu_device.model
    assert cpu_device.model == cpu_device.model.lower()
    for token in banned_tokens:
        assert token not in cpu_device.model
    assert isinstance(cpu_device.vendor, str) and cpu_device.vendor
    assert isinstance(cpu_device.features, list)
    assert all(isinstance(flag, str) and flag == flag.lower() for flag in cpu_device.features)


def test_cpu_memory_and_utilization(cpu_device):
    assert isinstance(cpu_device.memory_total, int) and cpu_device.memory_total > 0

    metrics = cpu_device.metrics()
    assert isinstance(metrics, CPUMetrics)
    assert 0 <= metrics.memory_used <= cpu_device.memory_total
    assert metrics.memory_process >= 0
    assert 0 <= metrics.utilization <= 100.0 + 1e-6

    mem_used_api = cpu_device.memory_used()
    util_api = cpu_device.utilization()
    allowed_diff = max(int(0.005 * cpu_device.memory_total), 64 * 1024 * 1024)
    assert abs(mem_used_api - metrics.memory_used) <= allowed_diff
    assert 0 <= util_api <= 100.0 + 1e-6


@pytest.mark.parametrize("fast", [False, True])
def test_cpu_metrics_paths(cpu_device, fast):
    metrics = cpu_device.metrics(fast=fast)
    assert isinstance(metrics, BaseMetrics)
    assert metrics.memory_used >= 0
    if metrics.memory_used:
        assert metrics.memory_used <= cpu_device.memory_total
    assert metrics.utilization >= 0



import time
import pytest

from device_smi.device import Device


class DummyDevice:
    """A minimal fake device that pretends to provide metrics."""

    fast_metrics_same_as_slow = False

    def __init__(self, parent):
        self.parent = parent
        self._counter = 0

    def metrics(self):
        self._counter += 1
        return type("Metrics", (), {
            "memory_used": 42,
            "utilization": 0.5,
        })()


def test_close_does_not_crash():
    """
    Ensure that calling close() stops the metrics thread cleanly
    without raising AttributeError: 'NoneType' object has no attribute 'wait'.
    """
    d = Device("cpu", fast_metrics_interval=0.05)
    # Patch the device with our dummy (to avoid depending on real hardware)
    d.device = DummyDevice(d)

    # Let the thread run at least once
    time.sleep(0.1)

    # Close repeatedly to simulate race conditions
    for _ in range(5):
        d.close()
        time.sleep(0.05)

    # If we get here without exception, test passes
    assert True


def test_close_allows_metrics_reuse():
    """
    After closing, calling metrics() should still work (using slow path).
    """
    d = Device("cpu", fast_metrics_interval=0.05)
    d.device = DummyDevice(d)

    # Grab metrics normally
    m1 = d.metrics()
    assert m1.memory_used == 42
    assert isinstance(m1.utilization, float)

    # Close and check metrics still available
    d.close()
    m2 = d.metrics()
    assert m2.memory_used == 42



import os
import re

from .base import GPU, BaseMetrics, GPUDevice, Pcie, _run


class AMDGPUMetrics(BaseMetrics):
    pass

class AMDDevice(GPUDevice):
    fast_metrics_same_as_slow = False

    def __init__(self, cls, index):
        super().__init__(cls, index)
        self.gpu_id = self._get_gpu_id()

        try:
            args = ["amd-smi", "static", "--gpu", f"{self.gpu_id}"]

            result = self.to_dict(_run(args=args).lower())
            market_name = re.findall(r'\[(.*?)]', result["market_name"])
            if market_name:
                market_name = market_name[0]
            else:
                market_name = result["market_name"]
            model = market_name.split("/")[0].strip()
            total_memory= result["size"].removesuffix("mb").strip()
            pci_bus_id = result["bdf"]
            pcie_gen = result['pcie_interface_version'].removeprefix("gen").strip()
            pcie_width = result['max_pcie_width']
            driver = result["driver"]
            features = [result['target_graphics_version']]

            args = ["amd-smi", "firmware", "--gpu", f"{self.gpu_id}"]
            lines = _run(args=args).lower().splitlines()

            result = {}
            current_id = None
            for line in lines:
                line = line.strip()
                if line.startswith("fw_id:"):
                    current_id = line.split(":")[1].strip()
                elif line.startswith("fw_version:") and current_id:
                    result[current_id] = line.split(":")[1].strip()
            firmware = f"{result}"

            if model.lower().startswith("amd"):
                model = model[len("amd"):]

            cls.model = model.strip().lower()
            cls.memory_total = int(total_memory) * 1024 * 1024  # bytes
            cls.vendor = "amd"
            cls.features = features
            cls.pcie = Pcie(gen=int(pcie_gen), speed=int(pcie_width), id=pci_bus_id)
            cls.gpu = GPU(driver=driver, firmware=firmware)
        except FileNotFoundError:
            raise FileNotFoundError()
        except Exception as e:
            raise e

    def _get_gpu_id(self):
        hips = os.environ.get("HIP_VISIBLE_DEVICES", "")
        hip_list = hips.split(",") if hips else []
        if hip_list and len(hip_list) > self.index:
            return hip_list[self.index]
        else:
            return str(self.index)

    def metrics(self):
        try:
            args = ["amd-smi", f"--id={self.gpu_id}", "--query-gpu=memory.used,utilization.gpu",
                    "--format=csv,noheader,nounits"]
            used_memory, utilization = _run(args=args, seperator="\n")[0].split(", ")

            return AMDGPUMetrics(
                memory_used=int(used_memory) * 1024 * 1024,  # bytes
                memory_process=0,  # Bytes, TODO, get this
                utilization=float(utilization),
            )

        except FileNotFoundError:
            raise FileNotFoundError(
                "The 'amd-smi' command was not found. Please ensure that the 'amd-utils' package is installed."
            )
        except Exception as e:
            raise e
